/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* VTK output functions
 *
 * $Id: vtk.c,v 1.38 2011/07/28 08:59:04 zlb Exp $
 */

#define USE_VECTORS	1

#include "phg.h"
#include "io.h"

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>	/* PATH_MAX */
#include <ctype.h>

static BOOLEAN vtk_nonlinear_elements = TRUE;

enum {VTK_SINGLE = 0, VTK_DOUBLE = 1};
static const char *vtk_precision_list[] = {"single", "double", NULL};
static int vtk_precision = VTK_SINGLE;

#define VTK_PUT_FLOAT(a) {			\
    if (vtk_precision == VTK_SINGLE) {		\
	float x = (float)(a);			\
	BigEndianAppend(&x, sizeof(x), 1);	\
    }						\
    else {					\
	double x = (double)(a);			\
	BigEndianAppend(&x, sizeof(x), 1);	\
    }						\
}

/* buffer for broadcasting filename */
static const char *filename;
/* buffer for broadcasting list of DOF ids */
static USHORT *dof_ids = NULL;	/* last element for filename length */
static USHORT dof_ids_len = 0;

static unsigned char *buffer = NULL;
static size_t buffer_pos, buffer_size;

static void *
swap_endian(void *buff, size_t n)
{
#if PHG_BIGENDIAN
    return buff;
#else	/* PHG_BIGENDIAN */
    static unsigned char buff0[8];
    unsigned char *p0, *p1;

    if (n <= 1)
	return buff;

    assert(sizeof(buff0) >= n);

    p0 = buff0 + n;
    p1 = buff;
    while (p0 > buff0) *(--p0) = *(p1++);
    return buff0;
#endif	/* PHG_BIGENDIAN */
}

static void
bigendian_copy(void *buf0, size_t size, size_t count, size_t loc)
{
    size_t i;
    char *buf = buf0;

    assert(buffer != NULL && loc + size * count <= buffer_size);
    for (i = 0; i < count; i++) {
	memcpy(buffer + loc, swap_endian(buf, size), size);
	buf += size;
	loc += size;
    }
}

#define BigEndianAppend(buf, size, count) {			\
    size_t size0 = size, count0 = count;			\
    bigendian_copy(buf, size0, count0, buffer_pos);		\
    buffer_pos += size0 * count0;				\
}

static char *
parse_name(GRID *g, int no)
{
    static char name0[128 + 16];
    char name[128];
    char *p = name;
    const char *s;
    int i, n;

    s = phgDofIndex2Pointer(g, dof_ids[no + 1])->name;

    /* check for duplicate names */
    for (i = 0, n = 0; i < no; i++)
	if (!strcmp(s, phgDofIndex2Pointer(g, dof_ids[i + 1])->name))
	    n++;

    while (*s != '\0' && p - name < sizeof(name) - 1) {
	*p = isspace(*(const BYTE *)s) ? '_' : *s;
	p++;
	s++;
    }
    *p = '\0';

    if (n == 0)
	sprintf(name0, "%s", name);
    else
	sprintf(name0, "%s_%d", name, n);

    return name0;
}

void
phgExportVTK_(GRID *g)
/* Output grid in VTK unstructured grid format. */
{
    int k, a, dim;
    size_t size, vtk_float_size;
    INT i;
    INT nleaf, nvert, nedge;
    SIMPLEX *s;
    DOF *dof;
    BOOLEAN has_cell_data = FALSE, has_point_data = FALSE;
    FLOAT *values;
    static char fn[PATH_MAX + 1];
    unsigned char c, output_order;

    ParallelFunction(phgExportVTK_, TRUE);

    if (g == NULL || g->nprocs <= phgRank)
	return;

    vtk_float_size = (vtk_precision == VTK_SINGLE ?
				sizeof(float) : sizeof(double));

    nleaf = g->nleaf;
    nvert = g->nvert;
    nedge = g->nedge;

    assert((g->flags & VERT_FLAG));
    assert(g->rank != 0 || sizeof(a) == 4);	/* to be fixed later */

#if USE_MPI
    if (phgMasterSlave) {
	if (g->rank == 0) {
	    memcpy(fn, filename, a = strlen(filename) + 1);
	    dof_ids[dof_ids_len - 1] = a ;
	}
	filename = fn;
	MPI_Bcast(&dof_ids_len, sizeof(dof_ids_len), MPI_BYTE, 0, g->comm);
	MPI_Bcast(dof_ids, sizeof(*dof_ids)*dof_ids_len, MPI_BYTE, 0, g->comm);
	a = dof_ids[dof_ids_len - 1];
	MPI_Bcast((void *)filename, a, MPI_BYTE, 0, g->comm);
    }
    else
#else
    {
	strcpy(fn, filename);
	filename = fn;
    }
#endif

    filename = phgIOAddExtension(filename, "vtk");
    if (!phgIOOpen(g, (char *)filename)) {
	phgError(0, "cannot create VTK file \"%s\".\n", filename);
	return;
    }
    if (g->rank == 0)
	phgInfo(1, "creating VTK file \"%s\".\n", filename);

    /* determine output order */
    if (vtk_nonlinear_elements == FALSE) {
	output_order = 1;
    }
    else {
	output_order = 0;
	for (a = 0; a < dof_ids[0]; a++) {
	    dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	    if (DofIsHP(dof)) {
		if (output_order < dof->hp->max_order)
		    output_order = dof->hp->max_order;
	    }
	    else {
		if (output_order < dof->type->order)
		    output_order = dof->type->order;
	    }
	}
    }

    if (output_order > 1 && g->period != NULL) {
	static BOOLEAN warned = FALSE;
	if (warned == FALSE && g->rank == 0)
	    phgWarning("nonlinear element for periodic boundaries unimplemented yet, using linear element instead.\n");
	warned = TRUE;
	output_order = 1;
    }

    if (output_order <= 1) {
	/* use linear tetrahedra */
	output_order = 1;
	buffer_size = nvert * vtk_float_size * 3;
	if ((size = nleaf * 5 * sizeof(a)) > buffer_size)
	    buffer_size = size;
    }
    else {
	/* use quadratic tetrahedra */
	output_order = 2;
	buffer_size = (nvert > nedge ? nvert : nedge) * vtk_float_size * 3;
	if ((size = nleaf * 11 * sizeof(a)) > buffer_size)
	    buffer_size = size;
    }

    /* compute size of DOF data */
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
        dim = DofDim(dof);
	dim = dim > 4 ? 4 : dim;
	if (dof->type == DOF_P0 || dof->type == DOF_DG0) {
	    size = nleaf * dim * vtk_float_size;
	    has_cell_data = TRUE;
	}
	else {
	    if (output_order <= 1)
		size = nvert * dim * vtk_float_size;
	    else
		size = (nvert + nedge) * dim * vtk_float_size;
	    has_point_data = TRUE;
	    if (DofIsHP(dof)) {
		has_cell_data = TRUE;	/* hp orders are saved as cell data */
		if (buffer_size < size)
		    buffer_size = size;
		size = nleaf * sizeof(unsigned char);
	    }
	}
	if (buffer_size < size)
	    buffer_size = size;
    }
    buffer = phgAlloc(buffer_size);

    /* File header and vertices */
    phgIOPrintf(g, "# vtk DataFile Version 2.0\n"
		"Tetrahedral grid created by PHG %s.\n"
		"BINARY\nDATASET UNSTRUCTURED_GRID\n"
		"POINTS %d %s\n", PHG_VERSION,
		g->nvert_global + (output_order > 1 ? g->nedge_global : 0),
		vtk_precision == VTK_DOUBLE ? "double" : "float");
    buffer_pos = 0;
    for (i = 0; i < nvert; i++) {
	for (a = 0; a < 3; a++) {
	    VTK_PUT_FLOAT(g->verts[i][a])
	}
    }
#if USE_MPI
    phgIOWrite(g, buffer, vtk_float_size * 3, nvert, g->nvert_global,
		g->types_vert, g->L2Gmap_vert, TRUE);
#else
    phgIOWrite(g, buffer, vtk_float_size * 3, nvert, g->nvert_global,
		NULL, NULL, FALSE);
#endif

    if (output_order > 1) {
	/* save list of nodes at edge centers */
	VEF_MAP *vef = phgDofSetupVEFMap(g, NULL, EDGE_FLAG);
	assert(vef != NULL);
	buffer_pos = 0;
	for (i = 0; i < nedge; i++) {
	    COORD *p0, *p1;
	    s = vef->Emap[i];
	    if (s == NULL) {
		/* unreferenced edge */
		VTK_PUT_FLOAT(0.)
		VTK_PUT_FLOAT(0.)
		VTK_PUT_FLOAT(0.)
		continue;
	    }
	    k = vef->Eind[i];
	    p0 = g->verts + s->verts[GetEdgeVertex(k, 0)];
	    p1 = g->verts + s->verts[GetEdgeVertex(k, 1)];
	    for (a = 0; a < 3; a++) {
		VTK_PUT_FLOAT(((*p0)[a] + (*p1)[a]) * 0.5)
	    }
	}
#if USE_MPI
	phgIOWrite(g, buffer, vtk_float_size * 3, nedge, g->nedge_global,
		   g->types_edge, g->L2Gmap_edge, TRUE);
#else
	phgIOWrite(g, buffer, vtk_float_size * 3, nedge, g->nedge_global,
		   NULL, NULL, FALSE);
#endif
	phgDofFreeVEFMap(&vef);
    }
    
    /* tetrahedra */
    phgIOPrintf(g, "\nCELLS %d %d\n", g->nelem_global,
		g->nelem_global * (output_order > 1 ? 11 : 5));
    buffer_pos = 0;
    ForAllElements(g, s) {
	/* Note: nodes in a VTK_QUADRATIC_TETRA are defined as:
	 * 	v0, v1, v2, v3, e01, e12, e02, e03, e13, e23 */
	static int edge_order[] = {0, 3, 1, 2, 4, 5};
	a = (output_order > 1 ? 10 : 4);
	BigEndianAppend(&a, sizeof(a), 1);
	/* vertices */
	for (i = 0; i < 4; i++) {
	    a = GlobalVertex(g, s->verts[i]);
	    BigEndianAppend(&a, sizeof(a), 1);
	}
	if (output_order <= 1)
	    continue;
	/* edge centers */
	for (i = 0; i < 6; i++) {
	    a = GlobalEdge(g, s->edges[edge_order[i]]) + g->nvert_global;
	    BigEndianAppend(&a, sizeof(a), 1);
	}
    }
    phgIOWrite(g, buffer, sizeof(a) * (output_order > 1 ? 11 : 5), nleaf,
		g->nelem_global, NULL, NULL, FALSE); 

    phgIOPrintf(g, "\nCELL_TYPES %d\n", g->nelem_global);
    buffer_pos = 0;
    a = (output_order > 1 ? /* VTK_QUADRATIC_TETRA = */24 : /*VTK_TETRA = */10);
    for (i = 0; i < nleaf; i++)
	BigEndianAppend(&a, sizeof(a), 1);
    phgIOWrite(g, buffer, sizeof(a), nleaf, g->nelem_global, NULL, NULL, FALSE);

    if (has_point_data) {	/* point data */
	phgIOPrintf(g, "\nPOINT_DATA %d\n",
		g->nvert_global + (output_order > 1 ? g->nedge_global : 0));
	for (a = 0; a < dof_ids[0]; a++) {
	    DOF *dof0;
	    dof = dof0 = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	    if (dof->type == DOF_P0 || dof->type == DOF_DG0)
		continue;
	    if (dof->type != DOF_Pn[output_order])
		dof = phgDofCopy(dof0, NULL, DOF_Pn[output_order], dof0->name);
	    if (g->rank == 0)
		phgInfo(2, "writing DOF %d: \"%s\"\n", a, dof->name);
	    dim = dof->dim;
	    if (dim > 4) {
		if (g->rank == 0)
		    phgWarning("only saving the first 4 components of \"%s\"\n",
				dof->name);
		dim = 4;
	    }
	    if (USE_VECTORS && dim == Dim)
	        phgIOPrintf(g, "\nVECTORS %s %s\n", parse_name(g, a),
			    vtk_precision == VTK_DOUBLE ? "double" : "float");
	    else
	        phgIOPrintf(g, "\nSCALARS %s %s %d\nLOOKUP_TABLE default\n",
			    parse_name(g, a),
			    vtk_precision == VTK_DOUBLE ? "double" : "float",
			    dim);
	    /* collect vertex data */
	    buffer_pos = 0;
	    values = DofVertexData(dof, 0);
	    for (i = 0; i < dim * nvert; i++) {
		VTK_PUT_FLOAT(*(values++))
	    }
#if USE_MPI
	    phgIOWrite(g, buffer, vtk_float_size * dim, nvert,
			g->nvert_global, g->types_vert, g->L2Gmap_vert, TRUE); 
#else
	    phgIOWrite(g, buffer, vtk_float_size * dim, nvert,
			g->nvert_global, NULL, NULL, FALSE); 
#endif
	    if (output_order > 1) {
		/* collect edge data */
		buffer_pos = 0;
		values = DofEdgeData(dof, 0);
		for (i = 0; i < dim * nedge; i++) {
		    VTK_PUT_FLOAT(*(values++))
		}
#if USE_MPI
		phgIOWrite(g, buffer, vtk_float_size * dim, nedge,
			g->nedge_global, g->types_edge, g->L2Gmap_edge, TRUE); 
#else
		phgIOWrite(g, buffer, vtk_float_size * dim, nedge,
			g->nedge_global, NULL, NULL, FALSE); 
#endif
	    }
	    if (dof != dof0)
		phgDofFree(&dof);
	}
    }

    phgIOPrintf(g, "\nCELL_DATA %d\n", g->nelem_global);
    if (has_cell_data) {
	for (a = 0; a < dof_ids[0]; a++) {
	    dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);

	    if (DofIsHP(dof)) {
		/* variable order DOF, save orders as cell data */
		phgIOPrintf(g, "\nSCALARS %s_orders unsigned_char 1\n"
			   "LOOKUP_TABLE default\n", parse_name(g, a));
		buffer_pos = 0;
		ForAllElements(g, s) {
		    c = dof->hp->elem_order[s->index];
		    BigEndianAppend(&c, sizeof(c), 1);
		}
		phgIOWrite(g, buffer, sizeof(c), nleaf, g->nelem_global,
				NULL, NULL, FALSE); 
		continue;
	    }

	    if (dof->type != DOF_P0 && dof->type != DOF_DG0)
		continue;
	    if (g->rank == 0)
		phgInfo(2, "writing DOF %d: \"%s\"\n", a, dof->name);
	    dim = dof->dim;
	    if (dim > 4) {
		if (g->rank == 0)
		    phgWarning("only saving the first 4 components of \"%s\"\n",
				dof->name);
		dim = 4;
	    }
	    if (USE_VECTORS && dim == Dim)
	 	phgIOPrintf(g, "\nVECTORS %s %s\n", parse_name(g, a),
			    vtk_precision == VTK_DOUBLE ? "double" : "float");
	    else
	 	phgIOPrintf(g, "\nSCALARS %s %s %d\nLOOKUP_TABLE default\n",
			    parse_name(g, a),
			    vtk_precision == VTK_DOUBLE ? "double" : "float",
			    dim);
	    buffer_pos = 0;
	    ForAllElements(g, s) {
		values = DofElementData(dof, s->index);
		for (k = 0; k < dim; k++) {
		    VTK_PUT_FLOAT(*(values++))
		}
	    }
	    phgIOWrite(g, buffer, vtk_float_size * dim, nleaf,
			g->nelem_global, NULL, NULL, FALSE); 
	}
    }

    /* save regional mark as cell data */
    phgIOPrintf(g, "\nSCALARS region_mark int 1\nLOOKUP_TABLE default\n");
    buffer_pos = 0;
    ForAllElements(g, s) {
	a = s->region_mark;
	BigEndianAppend(&a, sizeof(a), 1);
    }
    phgIOWrite(g, buffer, sizeof(a), nleaf, g->nelem_global, NULL, NULL, FALSE);

    /* save submesh indices as cell data */
    phgIOPrintf(g, "\nSCALARS submesh_no int 1\nLOOKUP_TABLE default\n");
    buffer_pos = 0;
    ForAllElements(g, s) {
	a = g->rank;
	BigEndianAppend(&a, sizeof(a), 1);
    }
    phgIOWrite(g, buffer, sizeof(a), nleaf, g->nelem_global, NULL, NULL, FALSE);

    phgIOClose();

    phgFree(buffer);
    buffer = NULL;

    return;
}

const char *
phgExportVTKn(GRID *g, const char *fn, int ndof, DOF **dofs)
/* Outputs mesh in VTK format, returns the output filename.
 * The variable part of the arguments is a NULL terminated list of DOFs
 * to save to the VTK file */
{
    int i, j, k;
    DOF *dof;
    BTYPE *types_vert = NULL;

    FreeAtExit(dof_ids);

    assert(fn != NULL && Dim == 3);
    if (g == NULL)
	return NULL;

    /* output fn */
    filename = fn;
    /* initialize dof_ids */
    dof_ids_len = ndof + 2;
    phgFree(dof_ids);
    dof_ids = phgAlloc(sizeof(*dof_ids) * dof_ids_len);

    /* list of DOFs */
    for (i = 0, k = 0; i < ndof; i++) {
	if ((dof = dofs[i])->type == DOF_ANALYTIC) {
	    if (g->rank == 0) {
		phgWarning("DOF type 'DOF_ANALYTIC' is not supported by "
			   "phgExportVTK(). \n");
		phgWarning("DOF '%s' will not be saved.\n", dof->name);
	    }
	    continue;
	}
	if (dof->g != g)
	    phgError(1, "%s: DOF \"%s\" doesn't belong to the grid.\n",
			__func__, dof->name);
	/* test for missing NULL terminator */
	if (!phgDofIsValid(g, dof)) {
	    if (g->rank == 0)
		phgWarning("phgExportVTK: missing NULL terminator"
			   " in argument list.\n");
	    break;
	}
	if ((j = phgDofPointer2Index(dof)) < 0)
	    break;
	dof_ids[++k] = j;
    }
    dof_ids[0] = k;

    if (g->period != NULL) {
	/* reconstruct types_vert[] without periodicity */
	BYTE	flags;
	PERIOD	*period;
	INT	nvert_owned, *owner_index_vert;
	SHORT	*owner_rank_vert;

	period = g->period;
	g->period = NULL;
	types_vert = g->types_vert;
	g->types_vert = NULL;
	owner_index_vert = g->owner_index_vert;
	g->owner_index_vert = NULL;
	owner_rank_vert = g->owner_rank_vert;
	g->owner_rank_vert = NULL;
	flags = g->flags;
	g->flags = VERT_FLAG;
	nvert_owned = g->nvert_owned;

	phgUpdateBoundaryTypes(g);

	g->period = period;
	phgFree(g->owner_index_vert);
	g->owner_index_vert = owner_index_vert;
	phgFree(g->owner_rank_vert);
	g->owner_rank_vert = owner_rank_vert;
	g->flags = flags;
	g->nvert_owned = nvert_owned;
    }

    phgExportVTK_(g);

    if (g->period != NULL) {
	phgFree(g->types_vert);
	g->types_vert = types_vert;
    }

    return filename;
}

const char *
phgExportVTK(GRID *g, const char *fn, DOF *dof, ...)
{
    int ndof;
    DOF **dofs;
    va_list ap;

    assert(g != NULL || phgInitialized == FALSE);

    if (g == NULL) {
	/* register options for VTK export */
	phgOptionsRegisterKeyword("-vtk_precision", "VTK data precision",
				  vtk_precision_list, &vtk_precision);
	phgOptionsRegisterNoArg("-vtk_nonlinear_elements",
		"Use VTK's non linear elements for high order elements",
		&vtk_nonlinear_elements);

	return NULL;
    }

    dofs = phgAlloc(256 * sizeof(*dofs));

    va_start(ap, dof);
    for (ndof = 0; ndof < 256; ndof++) {
	if (dof == NULL)
	    break;
	dofs[ndof] = dof;
	dof = va_arg(ap, DOF *);
    }
    va_end(ap);

    fn = phgExportVTKn(g, fn, ndof, dofs);
    phgFree(dofs);

    return fn;
}
